#!/usr/bin/bash

# call_graph.sh -- make a call graph of the FSCU software

# ----------------------------------------------------------------
# This file is part of code_graphs Copyright (C) 2011 Elijah 
# Laboratories Inc.  http://www.elilabs.com  ALL RIGHTS RESERVED.
#
# GNU cflow is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# code_graphs is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with code_graphs; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
# ----------------------------------------------------------------


echo '---------------- call graphs ----------------'

# default to local mode
MODE=${MODE:-"file:"}
if [ "$MODE" != "file:" ]
then MODE=${MODE}/call_graph/
fi

# get rid of any old results
rm -rf call_graph
mkdir -p call_graph

# start with a clean slate
rm -f *.pdf *.ps *.dot

# generate the call tree in text outline format
cflow --omit-symbol-names --omit-arguments -i s *.[ch] \
2> call_graph.err \
> call_graph.out

# convert to dot dependency lines
cat call_graph.out | ./cflow2dep.pl > call_graph.dup

# sort the list and purge duplicates
cat call_graph.dup | sort -u > call_graph.dot

# make all source nodes red trapezoids
cat call_graph.dot \
 | ./find_source_nodes.sh \
 | awk -v mode=$MODE -v dq='"' '{ print dq $0 dq " [shape=invtrapezium,style=filled,fillcolor=red,href=" dq mode  $1 ".pdf" dq "];" }' - \
> call_graph.sty

# find all widow nodes.  These are neither callers nor callees, but are defined.  Make them purple trapeziods.
cat call_graph.out \
 | grep '()' \
 | awk '{print $1}' - \
 | tr -d '()' \
 | sort -u \
> all_nodes.dat
cat find_source_nodes.lhs find_source_nodes.rhs \
 | sort -u \
 | join -v 2 - all_nodes.dat \
 | awk -v mode=$MODE -v dq='"' '{ print dq $0 dq " [shape=invtrapezium,style=filled,fillcolor=purple,href=" dq mode  $1 ".pdf" dq "];" }' - \
>> call_graph.sty

# make all sink nodes green rectangles
cat call_graph.dot \
 | ./find_sink_nodes.sh \
 | awk -v mode=$MODE -v dq='"' '{ print dq $0 dq " [shape=rectangle,style=filled,fillcolor=green,href=" dq mode  $1 ".pdf" dq "];" }' - \
>> call_graph.sty

# first, make a list of the source and sink nodes
cat call_graph.sty | tr -d '"' | awk '{ print $1 }' - | sort -u > source_and_sink.nodes

# next make a list of all nodes in the graph
cat call_graph.dot | ./all_nodes_in.sh > all.nodes

# now make a list of all non-source/sink nodes
join -v 1 all.nodes source_and_sink.nodes > other.nodes

# make style lines for them too as yellow ellipses
cat other.nodes \
 | awk -v mode=$MODE -v dq='"' '{ print dq $0 dq " [style=filled,fillcolor=yellow,href=" dq mode  $1 ".pdf" dq "];" }' - \
>> call_graph.sty

# determine stack frame size for each function
./stack_frames.sh > stack_frames.1

# annotate each node with module name and stack frame size
cat call_graph.sty | sort -u > stack_frames.2
join stack_frames.1 stack_frames.2 > stack_frames.3
cat stack_frames.3 \
 | awk -v dq='"' '{ print $1 " [label=" substr($1,1,length($1)-1) "\\" "n" $2,$3 dq "," substr($4,2) }' - > stack_frames.4
join -a 1 stack_frames.2 stack_frames.4 > stack_frames.5
cat stack_frames.5 \
 | awk 'NF == 2 { print $1,"[" substr($2,index($2,"href=")) ; next }
        NF == 4 { print $1,$3,$4 ; next }
                { print "---------------->>>>>>>> NF =",NF,"ERROR:", $0 }' - > stack_frames.6
cp stack_frames.6 call_graph.sty

# find any cycles in the overall call graph
cat call_graph.dot call_graph.sty > call_graph.cyc
./show_cycles.sh call_graph.cyc

# If there was recursion, show it; otherwise, graph the max stack critical path.

if [ -e "==CYCLES==.pdf" ]
then
     # Cycles were present, indicating recursion; stack analysis is not possible in this case.

     # re-color the arcs
     cat call_graph.dot | sort | awk '{ print $0 "|" }' > call_graph.1
     cat ==CYCLES==.dot | tr -d '"' | sort | sed 's/\ \[/\|\ \[/g' > call_graph.2
     join -t '|' -a 1 call_graph.1 call_graph.2 | tr -d '|' > call_graph.dot

     # re-color the nodes
     cat call_graph.sty | sort > call_graph.4
     cat ==CYCLES==.sty | sort > call_graph.5
     join -t '[' -a 1 -o '2.1 2.2 1.1 1.2' call_graph.4 call_graph.5 \
      | awk 'substr($0,1,2) == "[[" { print substr($0,3) ; next }
                                    { print substr($0,1,index($0,";")) ; next }' - > call_graph.sty

else
     # No cycles detected; perform stack analysis.

     # compute the max stack size and critical path to realize it
     ./stack_analysis.pl call_graph.dot call_graph.sty > stack_analysis.1

     # re-color the arcs
     cat call_graph.dot | sort | awk '{ print $0 "|" }' > call_graph.1
     awk '$2 == "->" { print $0,"[color=red,bold]" }' stack_analysis.1 | tr -d '";' | sort | sed 's/\ \[/\|\ \[/g' \
      | join -t '|' -a 1 call_graph.1 - | tr -d '|' > call_graph.dot

     # re-annotate the nodes to include maximum cumulative stack size
     cat call_graph.sty | sort > call_graph.2
     grep -v '\->' stack_analysis.1 | sort \
      | join -a 1 - call_graph.2 \
      | awk -v dq='"' '/\[label=/ { print $1,$4,$2 ":" $3 substr($5,index($5,dq)) ; next }
                                  { print $1,"[label=" substr($1,1,length($1)-1) "\\" "n" $2 ":" $3 dq "," substr($4,2) }' - \
      > call_graph.sty

fi

# generate the input files for the graph
cat call_graph.dot > dep.dot
cat call_graph.sty > dep.sty

# generate the overall call graph
make -f dot_makefile clean > /dev/null
make -f dot_makefile 2>&1 > /dev/null

# save the results
mv dep.pdf '==ALL==.pdf'
mv dep.ps '==ALL==.ps'
mv dep.in '==ALL==.dot'

# generate all of the subgraphs
./call_subgraphs.sh

# now make the table of contents
./call_graph_dir.sh > dir.html

# put in final home
cp *.html *.dot *.ps *.pdf *.gif call_graph/

# fix up title for browser
cat index.html.skel | sed 's/Dependency Graph Browser/Call Graph/' > call_graph/index.html

