#!/usr/bin/perl

# stack_analysis.pl <precedence_file> <attributes_file> -- analyze worst case stack usage

# ----------------------------------------------------------------
# This file is part of code_graphs Copyright (C) 2011 Elijah 
# Laboratories Inc.  http://www.elilabs.com  ALL RIGHTS RESERVED.
#
# GNU cflow is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# code_graphs is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with code_graphs; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
# ----------------------------------------------------------------


# pass the parameters

my $precedence_file = shift;			  # contains "A --> B", meaning
						  # A calls B and conversly, B
						  # is called by A.

my $attribute_file = shift;			  # contains '"name"
						  # [attribute...];' includes
						  # stack frame size

# open these files

open(PRECEDENCE, $precedence_file) || die "Can't open $precedence_file!\n";
open(ATTRIBUTE, $attribute_file) || die "Can't open $attribute_file!\n";


my %node;					  # the structure containing the call tree and stack data

# The node is a hash table keyed by node name.  Each item in the hash
# is a list containing the node name, the size of the stack frame that
# that function allocates on the stack, the maximum cumulative stack
# depth to reach this function from main(), a list of functions that
# call this fnction, and a list of functions that this function calls.


# First read the attribute file and instantiate a node for each function with its stack frame size

while(<ATTRIBUTE>) {
  chomp;

  ($name, $attr1, $attr2) = split;		  # name is quoted as first field, attr1 is label, attr2 is rest
  $name =~ tr/"//d;				  # strip off quotes from name

  $frame = substr($attr2, 0, index($attr2,'"'));  # extract stack frame size
  $frame = 0 if $frame eq "";			  # default to 0 if no size given

  if ($name eq "main") {			  # if this is the root, then stack cumulative stack == the frame
    $stack = $frame;
  } else {
    $stack = 0;
  }

  @{$node{$name}} = ($name, $frame, $stack, [ ], [ ]); # instantiate a node for the line
}


# Next read the precedence relations file to see who calls who.

# This file has records of the form A -> B that indicate A calls B.
# This also means that B is called by A.  We link forwards and
# backwards using the preds array for predecessors, and the succs
# array for successors.  Each time we read one of these lines, we link
# both links.  When we are finished reading the file, we have a
# forward and backward linked call graph represented in the "node"
# hash.

while(<PRECEDENCE>) {
  chomp;

  ($from,$op,$to) = split;			  # $from is the caller, $to is the called

  ($name, $frame, $stack, $preds, $succs) = @{$node{$from}}; # get $from node
  push @{$succs}, $to;				  # say $from calls $to

  ($name, $frame, $stack, $preds, $succs) = @{$node{$to}}; # get $to node
  push @{$preds}, $from				  # say $to is called by $from
}

# Now we traverse the graph structure stored in %node to determine the
# maximum stack used to reach any node in the tree.  This traversal is
# done in Knuth's pre-order, which is a variant of depth-first.  As we
# work our way down the tree, we accumulate the total stack depth at
# each node by adding the associated function's stack frame size to
# the accumulation of the node above it.  If the accumulated value is
# less than or equal to the value already stored for that node, then
# we cut the traversal off and backtrack to the next node.  If the
# result is greater than the value already stored for that node, then
# we store the new accumulated value and keep driving downward until
# we either reach a leaf node, or a node whose accumulated stack depth
# is greater than what we just computed.

# At the end of this, we will have a tree that is annotated with the
# maximum stack depth to reach each node.  If we then examine all the
# leaf nodes and find the one with the maximum depth, then we will have
# the maximum stack usage for the software.
# If we start at that leaf node with the deepest stack, and work our
# way up the tree to main, always choosing that caller with the
# deepest stack, then we will have a critical path that maps to the
# calling sequence that exhibits that maximum stack depth usage.
# There can be more than one path with that same maximum stack depth,
# in other words, multiple critical paths.

sub traverse {					  # traverse the tree in pre-order
  my $parent = shift;				  # the name of the root node of this subtree
  my $parent_name, $parent_frame, $parent_stack, $parent_preds, $parent_succs;

  my $child;
  my $child_name, $child_frame, $child_stack, $child_preds, $child_succs;

  my $new_stack;

  # get the root node to get the successors list
  ($parent_name, $parent_frame, $parent_stack, $parent_preds, $parent_succs) = @{$node{$parent}};

  foreach $child (@$parent_succs) {		  # for each child, if there are any

    # get its information
    ($child_name, $child_frame, $child_stack, $child_preds, $child_succs) = @{$node{$child}};

    $new_stack = $node{$parent}[2] + $child_frame;

    if ($new_stack > $child_stack) {		  # do we need to update the accumulated stack depth?
      $node{$child}[2] = $new_stack;		  # yes, do so
    }
    traverse($child);				  # continue to process child's children
  }
}

traverse("main");				  # traverse the tree and annotate with stack depth

foreach $routine (keys(%node)) {		  # print the hash table to stdout

  ($name, $frame, $stack, $preds, $succs) = @{$node{$routine}};

  print "\"$name\" $frame $stack\n";
}

# Now we need to determine the leaf node with the maximal stack depth.
# If we climb up the tree, always taking that node with maximal stack
# depth, we will be traversing the calling sequence that produces the
# maximal stack depth.  This is te critical path of the call graph
# with respect to stack depth.  We want to identify this path by
# coloring the edges along it differently from the default.

my @leaf;
my $max_depth = 0;

foreach $routine (keys(%node)) {		  # determine all te leaf nodes

  ($name, $frame, $stack, $preds, $succs) = @{$node{$routine}};

  push @leaf, $routine if @$succs == 0;		  # remember leaf nodes

  if ($stack > $max_depth) {			  # is this one deeper?
    $max_depth = $stack;			  # yes, update max depth so far
    $max_leaf = $routine;			  # remember where it was
  }
}
$routine = $max_leaf;

# Now that we have the leaf node, we want to output a sequence of
# edges from there to the root, which is a node with no predecessors.

for (;;) {
  ($name, $frame, $stack, $preds, $succs) = @{$node{$routine}};

  $npreds = @{$node{$routine}[3]};		  # how many predeecessors does this node have?
  last if $npreds == 0;				  # quit if no predecessors, we're at the root!

  $max_depth = -1;
  foreach $parent (@$preds) {			  # look at all of this node's predecessors

    $p_stack = $node{$parent}[2];		  # get parent's stack size
    if ($p_stack > $max_depth) {		  # is this one deeper?
      $max_depth = $p_stack;			  # yes, update max depth so far
      $max_pred = $parent;			  # remember where it was
    }
  }
  print "$max_pred -> $routine;\n";		  # output an edge on the critical path

  $routine = $max_pred;				  # move up to predecessor
}

